
import numpy as np
import scipy as sp

import matplotlib as mpl
from matplotlib.figure import Figure
from matplotlib.backends.backend_qtagg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.backends.backend_qtagg import NavigationToolbar2QT as NavigationToolbar

from PyQt6.QtCore import Qt
from PyQt6.QtWidgets import *

import DNMR.fileops as fileops
from DNMR.miniwidgets import *

from DNMR.tab import Tab

class TabInvLaplace(Tab):
    output_frames = {}

    def __init__(self, data_widgets, parent=None):
        super(TabInvLaplace, self).__init__(data_widgets, 'tab_inv_laplace', parent)
        
        self.plotted_data = []
        
    def generate_layout(self):
        l = QHBoxLayout()
        p = QPushButton('Fit')
        p.clicked.connect(self.fit)
        l.addWidget(p)
        return l

    def plot_logic(self):
        for i in self.plotted_data:
            
            self.ax.plot(i[0], i[1], label=i[2], alpha=0.5)
            
    def fit(self):
        ts = self.data_widgets['tab_phase'].data[0]
        freq = self.data_widgets['tab_ft'].data[0]
        ft   = self.data_widgets['tab_ft'].data[1]
        imag = np.imag(ft)
        real = np.real(ft)
        F = real + 1j*imag
        
        # for 7/2 spin.
        qs = np.array([1,6,15,28])
        ps = np.array([1/84, 3/44, 75/364, 1225/1716])
        num_bins = 250
        T1s = np.exp(np.linspace(np.log(4.5e5), np.log(5.5e5), num_bins))

        integrations = np.zeros(real.shape[0])
        start_index = np.argmin(np.abs(self.data_widgets['tab_ft'].left_pivot - freq))
        end_index = np.argmin(np.abs(self.data_widgets['tab_ft'].right_pivot - freq))
        if(end_index < start_index):
            tmp = start_index
            start_index = end_index
            end_index = tmp

        integrations = np.sum(F[:,start_index:end_index], axis=1)
        integrations -= np.min(integrations)
        integrations /= np.max(integrations)
        integrations *= 2.0
        integrations -= 1.0
        try:
            ts = self.fileselector.data.sequence['0'].delay_time
        except:
            ts = self.fileselector.data.sequence['0'].relaxation_time # Legacy
        
        inv_T1s = 1/T1s
        
        kernel = np.sum(1 - 2*ps[:,None,None]*np.exp(-qs[:,None,None] * ts[None,:,None]/T1s[None,None,:]), axis=0) # K[i,j]
        kernel = np.matrix(kernel)
        
        kernel = np.diag(np.linalg.svd(kernel)[1])
        kernel = np.resize(kernel, (len(ts), len(T1s)))
        
        def gaussian(x, sigma):
            g = np.exp(-1/2 * np.square((T1s - x)/sigma))
            return g/np.maximum(1e-9, np.sum(g))
        
        def cost_function(M, K, P, alpha):
            P = np.abs(P)
            return np.square(np.linalg.norm(M - K@P)) + alpha*np.square(np.linalg.norm(P))
            
        bounds = [ [ 1e-9, 1.0] for i in range(num_bins) ]
        # WAY UNDERDETERMINED
        self.plotted_data = []
        for a in [1e-1, 1e0, 1e1]:
            #res = sp.optimize.differential_evolution(lambda x, *args: cost_function(args[0], args[1], x, a), bounds, args=(integrations, kernel), constraints=(sp.optimize.LinearConstraint(np.identity(num_bins), 1.0, 1.01),))
            #x0 = np.ones(num_bins)/num_bins#np.exp(-np.square(np.linspace(-10, 10, num_bins))/2)
            #res = sp.optimize.minimize(lambda x, *args: cost_function(args[0], args[1], x, a), x0=x0, args=(integrations, kernel), method='SLSQP', constraints=({'type': 'eq', 'fun': lambda x: 1-np.sum(np.abs(x))},), options={'ftol':1e-9, 'maxiter': 2500})
            x0 = sp.optimize.brute(lambda x, *args: cost_function(args[0], args[1], gaussian(x, 10), a), ranges=[(T1s[0], T1s[-1])], Ns=len(T1s), args=(integrations, kernel), finish=None)
            sigma = sp.optimize.brute(lambda x, *args: cost_function(args[0], args[1], gaussian(x0, np.exp(x)), a), ranges=[(1, np.log(T1s[-1]))], Ns=len(T1s), args=(integrations, kernel), finish=None)
            sigma = np.exp(sigma)
            print(x0, sigma)
            
            P0 = gaussian(x0, sigma)
            res = sp.optimize.minimize(lambda x, *args: cost_function(args[0], args[1], x, a), x0=P0, args=(integrations, kernel), method='SLSQP', constraints=({'type': 'eq', 'fun': lambda x: 1-np.sum(np.abs(x))},))
            print(res)
            
            res_x = np.abs(res.x)
            normed = res_x / np.sum(res_x)
            self.plotted_data += [(T1s, normed, f'alpha={a}')]
            self.plotted_data += [(T1s, P0, f'G alpha={a}')]
        
        self.update()
        


import numpy as np
import h5py as hdf
import traceback
import re

def hdf_to_dict(g): # takes group, gives dict
    def t(n, g, d):
        ds = d
        for i in n.split('/')[:-1]:
            ds = ds[i]
        if(isinstance(g, hdf.Group)):
            ds[n.split('/')[-1]] = data_struct(hdf_to_dict(g))
        else:
            try:
                tmp = np.array(g)
            except:
                tmp = g
            ds[n.split('/')[-1]] = tmp
    d = {}
    g.visititems(lambda a,b: t(a,b,d))
    
    return d
    
class data_struct():
    data = None
    
    def __init__(self, init=None):
        self.data = {}
        
        if not(init is None):
            for k, v in init.items():
                self[k] = v
        
    def __getattr__(self, attr):
        try:
            return self.data[attr]
        except:
            return getattr(self.data, attr)
        
    def __getitem__(self, attr):
        return self.data[attr]
        
    def __setitem__(self, attr, val):
        if(isinstance(val, dict)):
            self.data[attr] = data_struct(val)
        else:
            self.data[attr] = val
        
    def __iadd__(self, r):
        for key in list(r.keys()):
            if not(key in self.data.keys()):
                self.data[key] = r[key]
                continue
            if(key == 'size'):
                self.data['size'] += r[key]
                continue
            val = self.data[key]
            if(isinstance(self.data[key], data_struct)):
                self.data[key] += r[key]
            else:
                try:
                    val = np.array(val)
                    setval = np.array(r[key])
                    if(val.ndim == 0):
                        val = np.array([val])
                    if(setval.ndim == 0):
                        setval = np.array([setval])
                    while(val.ndim < setval.ndim):
                        val = val[None,:]
                    self.data[key] = np.append(val, setval, axis=0) # if numpy
                except:
                    print(key, r[key])
                    self.data[key] += r[key] # they're lists if not.
        return self
        
    def __repr__(self):
        s = '(DATA_STRUCT) {\n'
        for key, val in self.data.items():
            if(isinstance(val, data_struct)):
                s += f'\t{key}: '
                s += '\n\t'.join(val.__repr__().split('\n')) + '\n'
            else:
                s += f'\t{key}: {val.__repr__()}\n'
        s += '}\n'
        return s



import numpy as np
import scipy as sp

import matplotlib as mpl
from matplotlib.figure import Figure
from matplotlib.backends.backend_qtagg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.backends.backend_qtagg import NavigationToolbar2QT as NavigationToolbar

from PyQt6.QtCore import Qt
from PyQt6.QtWidgets import *

from DNMR.miniwidgets import *
from DNMR.tab import Tab

class TabPhaseAdjustment(Tab):
    def __init__(self, data_widgets, parent=None):
        super(TabPhaseAdjustment, self).__init__(data_widgets, 'tab_phase', parent)
        
        self.data = (np.array([]), np.array([]))

    def generate_layout(self):
        self.phase_adjustment = PhaseAdjustmentWidget(callback=self.update_phase)
        self.data_widgets['phase_adjustment'] = self.phase_adjustment
        self.canvas.mpl_connect('button_press_event', self.process_button)
        
        self.spinbox_filtersize = QSpinBox()
        self.spinbox_filtersize.setRange(0, 100)
        self.spinbox_filtersize.setValue(12)
        self.spinbox_filtersize.valueChanged.connect(self.update)
        
        self.combobox_filtertype = QComboBox()
        self.combobox_filtertype.addItems(['Sinc','Gaussian','Half-Gaussian','Median','None'])
        self.combobox_filtertype.currentTextChanged.connect(self.update)
        
        self.pushbutton_phaseadjust = QPushButton('Autophase')
        self.pushbutton_phaseadjust.clicked.connect(self.autophase)
        
        self.pushbutton_applyall = QPushButton('Apply to all')
        self.pushbutton_applyall.clicked.connect(lambda: self.phase_set(self.phase_adjustment.slider_phase.value()))

        l2 = QVBoxLayout()
        l0 = QHBoxLayout()
        l0.addWidget(self.phase_adjustment)
        l3 = QVBoxLayout()
        l3.addWidget(self.pushbutton_phaseadjust)
        l3.addWidget(self.pushbutton_applyall)
        l0.addLayout(l3)
        l1 = QHBoxLayout()
        l1.addWidget(self.combobox_filtertype)
        l1.addWidget(self.spinbox_filtersize)
        l2.addLayout(l0)
        l2.addLayout(l1)
        return l2

    def autophase(self):
        if(self.data[0].shape[0] == 0):
            return
        
        # find peak, current phase at peak
        index = self.fileselector.spinbox_index.value()
        times = self.data[0][index]
        complexes = self.data[1][index]
        mags = np.abs(complexes)
        
        peak_i = np.argmax(mags)
        peak_t = times[peak_i]
        peak_c = complexes[peak_i]
        angle = np.arctan2(np.imag(peak_c),np.real(peak_c))
        angle_degrees = angle * 180.0/np.pi
        
        current_phases = self.get_global_phaseset() # degrees
        new_phase_degrees = current_phases[index]-angle_degrees
        # modulo 180
        new_phase_degrees += (np.abs(new_phase_degrees - 180.0)//360)*360.0
        new_phase_degrees -= (np.abs(new_phase_degrees + 180.0)//360)*360.0
        
        # now do the setting phase to zero.
        self.fileselector.data['phases'] = [ new_phase_degrees for i in range(len(self.fileselector.data['phases'])) ]
        self.get_global_peaklocs()
        self.fileselector.data['peak_locations'] = np.ones_like(self.get_global_peaklocs()) * peak_t
        
        self.update()

    def get_global_phaseset(self):
        if('phases' in self.fileselector.data.keys()):
            ps = self.fileselector.data['phases']
        else:
            self.fileselector.data['phases'] = [0 for i in range(self.fileselector.data['size'])]
            ps = self.fileselector.data['phases']
        return ps

    def get_global_peaklocs(self):
        if('peak_locations' in self.fileselector.data.keys()):
            ps = self.fileselector.data['peak_locations']
        else:
            self.fileselector.data['peak_locations'] = np.array([0 for i in range(self.fileselector.data['size'])])
            ps = self.fileselector.data['peak_locations']
        return ps
    
    def phase_set(self, p):
        ps = self.get_global_phaseset()
        for i in range(len(ps)):
            ps[i] = p
        self.update()

    def update_phase(self):
        ps = self.get_global_phaseset()
        index = self.fileselector.spinbox_index.value()
        ps[index] = self.phase_adjustment.slider_phase.value()
        self.update()

    def process_button(self, event):
        nav_state = self.ax.get_navigate_mode()
        if(nav_state is None):
            peak_locs = self.get_global_peaklocs() # Make sure it exists.
            if(event.button == 1):
                if not(event.xdata is None):
                    self.fileselector.data['peak_locations'] = np.ones_like(self.get_global_peaklocs()) * event.xdata
                self.update()
            elif(event.button == 3): # why set two or three indices, when you could set **just one**?
                index = self.fileselector.spinbox_index.value()
                if not(event.xdata is None):
                    self.fileselector.data['peak_locations'][index] = event.xdata
                self.update()

    def plot_logic(self):
        index = self.fileselector.spinbox_index.value()
        reals = self.fileselector.data.reals
        imags = self.fileselector.data.imags
        times = self.fileselector.data.times
        times = times[:,:reals.shape[1]]
        
        ps = self.get_global_phaseset()
        self.phase_adjustment.slider_phase.setValue(int(ps[index]))

        complexes = reals + 1j*imags
        
        for i in range(complexes.shape[0]):
            t = self.combobox_filtertype.currentText()
            if(t == 'Gaussian'):
                kernel = np.exp(-1/2 * np.square(np.linspace(-3, 3, self.spinbox_filtersize.value()*2+1)))
            elif(t == 'Sinc'):
                kernel = np.sinc(np.linspace(-reals.shape[1]/(2*self.spinbox_filtersize.value() + 1), reals.shape[1]/(2*self.spinbox_filtersize.value() + 1), reals.shape[1]))
            elif(t == 'Half-Gaussian'):
                i_s = np.linspace(-3, 3, self.spinbox_filtersize.value()*2+1)
                kernel = np.where(i_s >= 0, np.exp(-1/2 * np.square(i_s)), 0)
            elif(t == 'None'):
                kernel = 1
            elif(t == 'Median'):
                complexes[i] = sp.ndimage.median_filter(reals[i], mode='wrap', size=self.spinbox_filtersize.value()) + 1j * sp.ndimage.median_filter(imags[i], mode='wrap', size=self.spinbox_filtersize.value())
                continue
            kernel /= np.sum(kernel)
            complexes[i] = np.convolve(complexes[i], kernel, mode='same')

        phases = np.exp(1j*np.array(ps)* np.pi/180.0)
        complexes *= phases[:,None]

        #complexes -= np.average(complexes, axis=1)[:,None]

        peak_loc = self.get_global_peaklocs()[index]
        self.ax.axvline(peak_loc, color='k', linestyle='--', alpha=0.5)

        total = np.sum(np.square(np.abs(complexes[index]))) * (times[index][1] - times[index][0])

        reals = np.real(complexes)
        imags = np.imag(complexes)

        self.ax.plot(times[index][:reals[index].shape[0]], np.abs(complexes[index]), 'k', alpha=0.6, label=f'Mag.')
        self.ax.plot(times[index][:reals[index].shape[0]], reals[index], 'r', alpha=0.6, label='R')
        self.ax.plot(times[index][:imags[index].shape[0]], imags[index], 'b', alpha=0.6, label='I')

        self.ax.plot(times[index][:reals[index].shape[0]], np.average(reals, axis=0), 'r--', alpha=0.3, label='Avg. R')
        self.ax.plot(times[index][:imags[index].shape[0]], np.average(imags, axis=0), 'b--', alpha=0.3, label='Avg. I')
        self.ax.set_xlabel('time (us)')

        self.ax.axhline(0, color='k', linestyle='-')

        self.data = (times[:][:reals[index].shape[0]], complexes)
        
    def get_exported_data(self):
        index = self.fileselector.spinbox_index.value()
        d = { 'times': self.data[0][index],
                 'complexes': self.data[1][index],
                }
        return d